import os
import time
import zipfile
from sys import getsizeof

import numpy as np
import psutil
import torch as T
from PIL import Image


class DatasetStreamLoader:

    def __init__(self, file, batch_size, buffer_batches, shuffle=False):
        self.batch_size = batch_size
        self.buffer_batches = buffer_batches
        self.buffer_size = batch_size * buffer_batches
        self.shuffle = shuffle

        self.ptr = 0    # points into images and labels
        self.fin = zipfile.ZipFile(file, 'r')

        self.images_paths = sorted([path for path in self.fin.namelist(
        ) if path.startswith('Dataset/images/') and path.endswith('.jpg')])
        self.labels_paths = sorted([path for path in self.fin.namelist(
        ) if path.startswith('Dataset/labels/') and path.endswith('.txt')])

        # scale up dataset size for testing purposes: 10,000 * x
        # self.images_paths = [path for path in self.images_paths for i in range(100)]
        # self.labels_paths = [path for path in self.labels_paths for i in range(100)]

        self.images_paths_index = 0
        self.labels_paths_index = 0

        self.images = None
        self.labels = None

        # os.path.basename(self.images_paths[i]) # gets absolute path, useful
        # for matching label with image
        self.memory_message()
        self.reload_data()

    def memory_message(self):
        free_mem = psutil.virtual_memory().free
        # assumes all images in dataset has same dimensions
        example_image = np.asarray(Image.open(self.fin.open(
            self.images_paths[self.images_paths_index])))
        image_size = example_image.nbytes
        batch_mem = image_size * self.batch_size
        buffer_batch_mem = image_size * self.buffer_size

        print(
            f"\n\nWARNING: You have a total of {(free_mem / (1024 ** 2)):.2f} MB of memory "
            f"available on your machine. Each batch of {self.batch_size} images will "
            f"consume {(batch_mem / (1024 ** 2)):.2f} MB of memory. Each buffer "
            f"of {self.buffer_batches} batches (a total of {self.buffer_size} images) will consume "
            f"{(buffer_batch_mem / (1024 ** 2)):.2f} MB of memory. Please make sure "
            f"you have enough memory for at least one buffer, as memory is freed up "
            f"after each buffer is completed.\n\n")

    def reload_data(self):
        xy_lst = []
        ct = 0  # number of files read

        while ct < self.buffer_size:
            if self.images_paths_index >= len(self.images_paths):
                return False

            image_archived = self.fin.open(
                self.images_paths[self.images_paths_index])
            image = np.asarray(Image.open(image_archived))

            label_archived = self.fin.open(
                self.labels_paths[self.labels_paths_index])
            label = np.asarray([line.split() for line in label_archived.read().decode(
                'utf-8').splitlines()], dtype=np.float32)

            xy_lst.append((image, label))
            self.images_paths_index += 1
            self.labels_paths_index += 1
            ct += 1

        if self.shuffle:
            np.random.shuffle(xy_lst)

        images_mat = np.array([xy[0] for xy in xy_lst])
        labels_mat = np.array([xy[1] for xy in xy_lst])

        self.images = T.tensor(images_mat, dtype=T.uint8).to()
        self.labels = T.tensor(labels_mat, dtype=T.float32).to()

        return True

    def __iter__(self):
        return self

    def __next__(self):
        print('index: ', self.images_paths_index)
        if self.ptr + self.batch_size > self.buffer_size:
            self.ptr = 0
            res = self.reload_data()
            if not res:
                raise StopIteration

        start = self.ptr
        end = self.ptr + self.batch_size
        x = self.images[start:end]
        y = self.labels[start:end]
        self.ptr += self.batch_size
        return (x, y)


def main(batch_size, buffer_batches, dataset):
    np.random.seed(1)

    dataset_file = dataset

    dataloader = DatasetStreamLoader(
        dataset_file,
        batch_size,
        buffer_batches,
        shuffle=True)

    max_epochs = 1
    for epoch in range(max_epochs):
        print('epoch: ', epoch)
        for (index, batch) in enumerate(dataloader):
            print('batch: ', index)
            print('images: ', batch[0].shape)
            print('labels: ', batch[1].shape)
            print()


if __name__ == '__main__':
    start_time = time.time()

    # batch_size * buffer_batches should divide the number of images in the
    # dataset to avoid dropping data, and should be less than the number of
    # images in the dataset
    batch_size = 64
    buffer_batches = 16

    main(batch_size, buffer_batches)
    print('time elapsed: ', time.time() - start_time)



import json


def write_file(file_path: str, string: str):
    with open(file_path, "w") as file:
        file.write(string)


def read_file(file_path: str):
    with open(file_path, "r") as file:
        return file.read()


def write_json(file_path: str, data_structure):
    with open(file_path, "w") as file:
        json.dump(data_structure, file)


def read_json(file_path: str):
    with open(file_path, "r") as file:
        return json.load(file)



import os

import click

_INTERNAL_FOLDER_NAME = ".flockfysh"


def _find_flockfysh_workspace() -> str or None:
    prev_path = None
    cur_path = os.getcwd()
    while True:
        if cur_path == prev_path:
            cur_path = None
            break
        else:
            if os.path.isdir(os.path.join(cur_path, ".flockfysh")):
                break
            prev_path = cur_path
        cur_path = os.path.realpath(os.path.join(cur_path, ".."))
    return cur_path


_global_workspace_dir = _find_flockfysh_workspace()


def flockfysh_path(directory):
    global _global_workspace_dir
    return os.path.join(_global_workspace_dir, _INTERNAL_FOLDER_NAME, directory)


def _check_flockfysh_dir(func):
    def decorated(*args, **kwargs):
        if _global_workspace_dir is not None and os.path.exists(
                os.path.join(_global_workspace_dir, _INTERNAL_FOLDER_NAME)):
            return func(*args, **kwargs)
        else:
            raise FileNotFoundError("Flockfysh directory not found.")

    return decorated


def init_project():
    if _global_workspace_dir is not None:
        if click.confirm(
                f"Are you sure to create a new project? You might want to use another folder instead. {_global_workspace_dir}"):
            os.makedirs(_INTERNAL_FOLDER_NAME, exist_ok=True)
    os.makedirs(_INTERNAL_FOLDER_NAME, exist_ok=True)
    os.makedirs(os.path.join(_INTERNAL_FOLDER_NAME, "assets"), exist_ok=True)
    os.makedirs(os.path.join(_INTERNAL_FOLDER_NAME, "streams"), exist_ok=True)



import math
import os.path
import typing
import uuid
from ..internals.project import flockfysh_path
from ..internals.errors import NotFoundException
from ..internals.api import _api_session
from ..internals import file
from ordered_set import OrderedSet
from ..asset.image import Image
from ..asset.asset import Asset
import asyncio
from zipfile import ZipFile
import shutil


class Stream:
    _stream_store = {}

    @staticmethod
    def _get_stream_path(stream_id: str):
        return flockfysh_path(os.path.join("streams", stream_id))

    @staticmethod
    def _generate_stream_id():
        while True:
            new_id = uuid.uuid4().hex
            if not os.path.exists(flockfysh_path(os.path.join("streams", new_id))):
                return new_id

    @staticmethod
    def load(saved_stream_id: str):
        cached_stream = Stream._stream_store.get(saved_stream_id, None)
        if cached_stream is not None:
            return cached_stream
        if not os.path.exists(Stream._get_stream_path(saved_stream_id)):
            raise NotFoundException("This stream does not exist!")
        saved_stream_info_path = os.path.join(Stream._get_stream_path(saved_stream_id), "info.json")
        saved_stream_info = file.read_json(saved_stream_info_path)
        saved_stream = Stream(_internal=True,
                              _stream_id=saved_stream_id,
                              _dataset_id=saved_stream_info["dataset_id"],
                              _queries=saved_stream_info["queries"],
                              _dataset_info=saved_stream_info["dataset_info"])
        Stream._stream_store[saved_stream_id] = saved_stream
        return saved_stream

    @staticmethod
    def _create(dataset_id: str, dataset_info, *,
                asset_stage: typing.Literal["completed", "feedback", "uploaded"] = None,
                asset_filename_search: str = None):
        _queries = {
            "stage": asset_stage,
            "displayName": asset_filename_search,
        }
        new_stream = Stream(_internal=True, _stream_id=Stream._generate_stream_id(), _dataset_id=dataset_id,
                            _dataset_info=dataset_info, _queries=_queries)
        file.write_json(new_stream.__stream_info_path,
                        {"dataset_id": new_stream.__dataset_id, "queries": new_stream.__queries,
                         "dataset_info": new_stream.__dataset_info})
        Stream._stream_store[new_stream.stream_id] = new_stream
        return new_stream

    @property
    def _stream_seek(self):
        if self.__stream_seek is None:
            try:
                self.__stream_seek = file.read_json(self.__stream_seek_path)
                result = self.__stream_seek["stream_seek"]
                return result
            except (IOError, KeyError, TypeError):
                self.__stream_seek = {"stream_seek": {}}
                file.write_json(self.__stream_seek_path, self.__stream_seek)
        return self.__stream_seek["stream_seek"]

    @_stream_seek.setter
    def _stream_seek(self, seek_structure=None):
        self.__stream_seek = {"stream_seek": seek_structure}
        file.write_json(self.__stream_seek_path, self.__stream_seek)

    @property
    def _item_ids(self):
        if self.__item_ids is None:
            if os.path.exists(self.__item_ids_path):
                data = file.read_json(self.__item_ids_path)
                self.__item_ids = OrderedSet(data)
            else:
                self.__item_ids = OrderedSet()
                file.write_json(self.__item_ids_path, list(self.__item_ids))
        return self.__item_ids

    @_item_ids.setter
    def _item_ids(self, item_ids=None):
        if item_ids is None:
            item_ids = []
        self.__item_ids = list(item_ids)
        file.write_json(self.__item_ids_path, self.__item_ids)
    
    @property
    def _iterator_start_index(self):
        if self.__iterator_start_index is None:
            if os.path.exists(self.__iterator_start_index_path):
                self.__iterator_start_index = file.read_json(self.__iterator_start_index_path)
            else:
                self.__iterator_start_index = 0
                file.write_json(self.__iterator_start_index_path, self.__iterator_start_index)
        return self.__iterator_start_index

    @_iterator_start_index.setter
    def _iterator_start_index(self, new_index):
        if not isinstance(new_index, int):
            raise TypeError("Iterator start index must be an integer.")
        self.__iterator_start_index = new_index
        file.write_json(self.__iterator_start_index_path, self.__iterator_start_index)

    @property 
    def _cache_size(self):
        return len(self.__item_data_cache)

    @property
    def _finished(self):
        if self.__finished is None:
            if os.path.exists(self.__stream_finished_path):
                self.__finished = file.read_json(self.__stream_finished_path)
            else:
                self.__finished = False
                file.write_json(self.__stream_finished_path, self.__finished)
            return self.__finished
        return self.__finished

    @_finished.setter
    def _finished(self, new_state):
        if not isinstance(new_state, bool):
            raise TypeError("Finished state must be true or false.")
        self.__finished = new_state
        file.write_json(self.__stream_finished_path, self.__finished)

    @property
    def finished(self):
        return self._finished

    def __init__(self, *_args,
                 _stream_id: str,
                 _dataset_id: str = None,
                 _queries=None,
                 _internal=False,
                 _dataset_info):
        if _queries is None:
            _queries = {}
        if not _internal:
            raise RuntimeError("This class cannot be called by the user.")
        self.__stream_id = _stream_id
        self.__dataset_id = _dataset_id
        self.__dataset_info = _dataset_info
        self.__queries = _queries
        self.__stream_path = Stream._get_stream_path(self.__stream_id)
        os.makedirs(self.__stream_path, exist_ok=True)
        self.__stream_info_path = os.path.join(self.__stream_path, "info.json")
        self.__stream_seek_path = os.path.join(self.__stream_path, "seek.json")
        self.__stream_finished_path = os.path.join(self.__stream_path, "finished.json")
        self.__item_ids_path = os.path.join(self.__stream_path, "item_ids.json")
        self.__item_data_path = os.path.join(self.__stream_path, "item_data")
        self.__iterator_start_index_path = os.path.join(self.__stream_path, "iterator_start_index.json")
        os.makedirs(self.__item_data_path, exist_ok=True)
        self.__item_data_cache = {}
        self.__stream_seek = None
        self.__item_ids = None
        self.__finished = None
        self.__iterator_start_index = None

    @property
    def _dataset_info(self):
        return self.__dataset_info

    @property
    def classes(self):
        return self.__dataset_info["classes"]

    @property
    def stream_id(self):
        return self.__stream_id

    @property
    def stream_queries(self):
        return self.__queries

    @property
    def dataset_id(self):
        return self.__dataset_id
    
    def __getitem(self, item_id):
        try:
            item = self.__item_data_cache[item_id]
            return item
        except KeyError:
            try:
                item = file.read_json(os.path.join(self.__item_data_path, f"{item_id}.json"))
                item_type = item["type"]
                if item_type == "image":
                    abstract_item = Image(
                        asset_id=item["_id"],
                        asset_mimetype=item["mimetype"],
                        annotation_data=item["annotation_data"],
                        asset_stage=item["stage"],
                        asset_workspace_path=item["path"],
                        _internal=True)
                else:
                    abstract_item = Asset(
                        asset_id=item["_id"],
                        asset_stage=item["stage"],
                        asset_mimetype=item["mimetype"],
                        asset_workspace_path=item["path"],
                        _internal=True)
                self.__item_data_cache[item_id] = abstract_item
                return self.__item_data_cache[item_id]
            except IOError:
                raise RuntimeError("This item is not found.")
    
    def getitem(self, item_id):
        try:
            item = self.__item_data_cache[item_id]
            return item
        except KeyError:
            try:
                item = file.read_json(os.path.join(self.__item_data_path, f"{item_id}.json"))
                item_type = item["type"]
                if item_type == "image":
                    abstract_item = Image(
                        asset_id=item["_id"],
                        asset_mimetype=item["mimetype"],
                        annotation_data=item["annotation_data"],
                        asset_stage=item["stage"],
                        asset_workspace_path=item["path"],
                        _internal=True)
                else:
                    abstract_item = Asset(
                        asset_id=item["_id"],
                        asset_stage=item["stage"],
                        asset_mimetype=item["mimetype"],
                        asset_workspace_path=item["path"],
                        _internal=True)
                self.__item_data_cache[item_id] = abstract_item
                return self.__item_data_cache[item_id]
            except IOError:
                raise RuntimeError("This item is not found.")

    def __getitem__(self, item):
        if isinstance(item, str):
            return self.__getitem(item)
        elif isinstance(item, int):
            item_no = item
            self.expand(item_no + 1)
            item_id = self._item_ids[item_no]
            return self.__getitem(item_id)
        elif isinstance(item, slice):
            cur_slice = item
            start = cur_slice.start
            stop = cur_slice.stop
            step = cur_slice.step
            if start is None:
                start = 0
            if stop is None or stop < 0 or start < 0:
                self.expand()
            else:
                self.expand(max(start + 1, stop + 1))
            return list(map(lambda item_id: self.__getitem(item_id), self._item_ids[start:stop:step]))

    def _save_item(self, obj):
        to_save = obj._to_dict()
        self.__item_data_cache[obj.id] = obj
        file.write_json(os.path.join(self.__item_data_path, f"{obj.id}.json"), to_save)

    def __iter__(self):
        return self.create_iterator(pre_download=True)

    def create_iterator(self, *, start_index: int = 0, chunk_size: int = None, pre_download: bool = False,
                        batch_size: int = 20, reset=False, save=False):
        if chunk_size is not None:
            if not isinstance(chunk_size, int):
                raise AssertionError("Chunk size must be an integer.")
            if chunk_size <= 0:
                raise AssertionError("Chunk size must be greater than 0.")
        if chunk_size is not None:
            self.free()

        async def download_current_page(current_page):
            coroutines = []
            for item in current_page:
                item: Asset
                coroutines.append(item._download())
            await asyncio.gather(*coroutines)

        def process_current_page(current_page):
            if pre_download:
                asyncio.run(download_current_page(current_page))
            if save:
                self.download(whole=False, chunk=True, assets=current_page, reset=reset)
            self._iterator_start_index = len(self._item_ids)
            for item in current_page:
                yield item

        def purge_older_chunks(index):
            if chunk_size is not None:
                self.free(0, math.floor(index / chunk_size) * chunk_size)

        prev_len = start_index
        while True:
            current_page = []
            for i in range(prev_len, len(self._item_ids)):
                current_item_id = self._item_ids[i]
                item = self[current_item_id]
                if len(current_page) >= batch_size:
                    for item in process_current_page(current_page):
                        yield item
                    current_page = []
                    purge_older_chunks(i)
                current_page.append(item)
            for item in process_current_page(current_page):
                yield item
            purge_older_chunks(prev_len)
            prev_len = len(self._item_ids)
            try:
                self.next_assets(batch_size)
            except StopIteration:
                return 

    @property
    def dataset(self):
        from ..dataset.dataset import Dataset
        return Dataset(self.dataset_id)

    def next_assets(self, count: int = 20):
        if self.finished:
            raise StopIteration("Stream has ended.")
        
        response = _api_session.get(f'/api/datasets/{self.dataset_id}/assets', data={
            "stage": self.__queries.get("stage", None),
            "next": self._stream_seek.get("next", None),
            "limit": count
        })

        json_result = response.json()["data"]
                
        result = json_result["data"]
        cursor = json_result["meta"]
        if response.status_code >= 500:
            raise RuntimeError("There is a server error. Please try again later.")

        asset_objects = []

        async def process_response():
            # Download cycle
            for asset in result:
                asset_type = asset["type"]
                if asset_type == "image":
                    asset_obj = Image(
                        asset_id=asset["_id"],
                        annotation_data=asset["annotationData"],
                        asset_url=asset["url"],
                        asset_mimetype=asset["mimetype"],
                        asset_stage=asset["stage"],
                        _internal=True)
                else:
                    asset_obj = Asset(
                        asset_id=asset["_id"],
                        asset_url=asset["url"],
                        asset_stage=asset["stage"],
                        asset_mimetype=asset["mimetype"],
                        _internal=True)
                asset_objects.append(asset_obj)

        asyncio.run(process_response())

        for asset_object in asset_objects:
            self._save_item(asset_object)
            self._item_ids.append(asset_object.id)
        self._item_ids = self._item_ids

        if not cursor["hasNext"]:
            self._finished = True
        else:
            self._stream_seek = cursor

        return asset_objects

    def expand(self, length=None):
        while length is None or len(self._item_ids) < length:
            try:
                self.next_assets(20)
            except StopIteration:
                break

    def free(self, start=None, end=None, ignore_errors=False):
        for asset_id in self._item_ids[start:end]:
            asset = self.__getitem(asset_id)
            asset.free(ignore_errors=ignore_errors)
        self.__item_data_cache = {}

    def _construct_yolov5_yaml(self):
        names = "\n".join(f"  {index}: {class_name}" for index, class_name in enumerate(self.classes))
        output = f'''path: ../dataset
train: train/images
val: valid/images

names:
{names}
        '''
        return output

    def download(self, whole=False, chunk=False, assets=False, reset=False):
        save_location = os.path.join(os.getcwd(), os.path.join('downloads', self.stream_id))
        if not os.path.exists(os.path.join(os.getcwd(), 'downloads')):
            os.mkdir(os.path.join(os.getcwd(), 'downloads'))
        if not os.path.exists(save_location):
            os.mkdir(save_location)
        elif reset:
            shutil.rmtree(save_location)
            os.mkdir(save_location)

        if whole:
            self.expand()
        if not chunk:
            for asset in self:
                shutil.copy(asset.path, save_location)
        else:
            if not isinstance(assets, list):
                print('[ERROR] Assets must be a list.')
                return
            for asset in assets:
                shutil.copy(asset.path, save_location)
        
        print(f'\n[SUCCESS] Dataset saved to {save_location}.')

    # def to_yolov5_zip(self, save_location: str, confidence_level: float = 0.4):
    #     with ZipFile(save_location, "w", allowZip64=True) as new_zip:
    #         root_path = os.path.join("/", "dataset")
    #         asset: Image
    #         for asset in self:
    #             rand_output = random.random()
    #             if rand_output < confidence_level:
    #                 asset_type = "train"
    #             else:
    #                 asset_type = "valid"
    #             new_zip.write(asset.path, os.path.join(root_path, asset_type, "images", asset.filename))
    #             new_zip.writestr(os.path.join(root_path, asset_type, "labels", f'{asset.id}.txt'),
    #                              asset.pytorch_annotation_string)
    #         new_zip.writestr(os.path.join(root_path, "dataset.yaml"), self._construct_yolov5_yaml())
    #
    # def to_yolov5_model(self, save_location: str, confidence_level: float = 0.4):
    #     shutil.rmtree(save_location, ignore_errors=True)
    #     root_path = os.path.join(save_location, "dataset")
    #     os.makedirs(root_path)
    #     asset: Image
    #     for asset in self:
    #         rand_output = random.random()
    #         if rand_output < confidence_level:
    #             asset_type = "train"
    #         else:
    #             asset_type = "valid"
    #         image_path = os.path.join(root_path, asset_type, "images")
    #         label_path = os.path.join(root_path, asset_type, "labels")
    #         os.makedirs(image_path, exist_ok=True)
    #         os.makedirs(label_path, exist_ok=True)
    #         shutil.copy2(asset.path, os.path.join(image_path, asset.filename))
    #         with open(os.path.join(label_path, f'{asset.id}.txt'), "w",
    #                   encoding="UTF-8") as file:
    #             file.write(asset.pytorch_annotation_string)
    #     with open(os.path.join(root_path, 'dataset.yaml'), 'w') as file:
    #         file.write(self._construct_yolov5_yaml())



import copy
import pprint
import typing

from ..internals.api import _api_session
from ..internals import project, api, errors
from ..stream.stream import Stream


class Dataset:
    @project._check_flockfysh_dir
    @_api_session._authorization_check
    def __init__(self, dataset_id):
        self.__dataset_id = dataset_id

        result = _api_session.get(f"/api/datasets/stats/{dataset_id}", data={
            "expand": "assetCounts",
        }).json()

        if not result["success"]:
            error_code = result["error"]["code"]
            if error_code == "ERROR_NOT_FOUND":
                raise errors.NotFoundException("Dataset not found.")
            else:
                raise RuntimeError("There is a problem fetching datasets")
        self.__dataset_info = result["data"]

    @property
    def info(self):
        return copy.deepcopy(self.__dataset_info)

    @property
    def dataset_id(self):
        return self.__dataset_id
    
    @property
    def length(self):
        return self.__dataset_info["assetCounts"]["total"]

    def create_stream(self, *, asset_stage: typing.Literal["completed", "uploaded", "feedback"] = None,
                      asset_filename_search: str = None):
        return Stream._create(self.__dataset_id, copy.deepcopy(self.__dataset_info), asset_stage=asset_stage,
                              asset_filename_search=asset_filename_search)



import json
import time

from . import project
from .api import _api_session
import webbrowser
from getpass import getpass

def get_oauth_instance():

    apiKey = getpass("Enter your flockfysh client id. This can be found in the 'Settings' page of your profile: ")
    
    print(str(apiKey))
    
    new_instance = _api_session.post("/api/users/auth/register", data={
        "scope": "write_dataset read_dataset share_dataset manage_account",
        "apiKey": str(apiKey),
    }).json()['data']
    
    return new_instance["data"]


def poll_oauth_instance(deviceCode):
    while True:
        time.sleep(2)
        private_instance = _api_session.post("/api/users/auth/getAccessToken/private", data={
            "deviceCode": deviceCode
        }).json()["data"]["data"]


        state = private_instance['state']
        if state == "approved":
            return private_instance["access_token"]["access_token"]
        elif state == "rejected":
            raise Exception("Login cancelled.")
        elif state == "expired":
            raise Exception("Login session expired. Try logging in again.")


@project._check_flockfysh_dir
def login(open_window=True):
    new_instance = get_oauth_instance()

    print(new_instance)
    if open_window:
        webbrowser.open(new_instance["url"])
    else:
        print(f"Go to {new_instance['url']} to log in to the CLI.")

    print(new_instance['device_code'])
    access_token = poll_oauth_instance(new_instance["device_code"])
    with open(project.flockfysh_path("credentials.json"), "w") as file:
        json.dump({
            "access_token": access_token
        }, file)



import click
from .modules.internals import project, authorization
from . import Stream, Dataset


@click.group()
def ffysh():
    pass


@ffysh.command()
def init():
    project.init_project()


@ffysh.command()
def login():
    authorization.login()


@ffysh.command()
@click.option('-d', '--dataset', help='Enter dataset ID which can be located in a dataset\'s URL.', required=True)
def download(dataset):
    """Downloads an entire dataset."""
    try:
        stream = Dataset(dataset_id=dataset).create_stream()
    except:
        click.echo("Dataset either is private or does not exist.")
        return
    stream.download(whole=True)


@ffysh.command()
@click.option('-d', '--dataset', help='Enter dataset ID which can be located in a dataset\'s URL.', required=True)
@click.option('-b', '--batch', type=int, help='Specifies the number of files to be downloaded. Starts off where specified stream ended or first file if stream is not specified', required=True)
@click.option('-c', '--chunk', type=int, help='The number of filed stored locally is equivalent to the chunk size plus the batch size.', required=True)
@click.option('-s', '--stream', help='Enter stream ID to continue chunking from a previous session.', required=False)
@click.option('-np', '--nopredownload', is_flag=True, help="Blocks downloading multiple files at once when iterating through the stream.")
@click.option('-r', '--reset', is_flag=True, help="Deletes all files (previous chunks) in saved stream directory")
@click.option('-ns', '--nosave', is_flag=True, help='Does not save stream to a directory. Assets can be found in the assets folder of the project directory.')
def load(dataset, batch, chunk, stream, nopredownload=False, reset=False, nosave=False):
    dataset_id = dataset
    batch_size = batch
    chunk_size = chunk
    stream_id = stream
    save = not nosave
    predownload = not nopredownload
    """Loads the next batch of a dataset."""
    if not isinstance(batch_size, int):
        raise AssertionError("Batch size must be an integer.")
    if batch_size <= 0:
        raise AssertionError("Batch size must be greater than 0.")
    if not isinstance(chunk_size, int):
        raise AssertionError("Chunk size must be an integer.")
    if chunk_size <= 0:
        raise AssertionError("Chunk size must be greater than 0.")

    try:
        dataset = Dataset(dataset_id=dataset_id)
    except:
        click.echo("Dataset either is private or does not exist.")
        return

    if stream is not None:
        try: 
            stream = Stream.load(saved_stream_id=stream_id)
        except:
            click.echo("Stream does not exist.")
            return
        if stream.dataset_id != dataset_id:
            click.echo(f"Stream belongs to incorrect dataset. Stream belongs to dataset {stream.dataset_id} while dataset {dataset_id} was specified.")
            return
        it = stream.create_iterator(start_index=stream._iterator_start_index, chunk_size=chunk_size, pre_download=predownload, batch_size=batch_size, save=save, reset=reset)
    else:
        stream = dataset.create_stream()
        click.echo(f"Created stream {stream.stream_id}. Use this stream ID to continue chunking from a previous session.")
        it = stream.create_iterator(start_index=0, chunk_size=chunk_size, pre_download=predownload, batch_size=batch_size, save=save, reset=reset)

    try:
        next(it)
    except StopIteration:
        click.echo("Stream already finished.")
        return
    
# @click.command()
# @click.option('-b', '--batch', default=64, help='# images per batch')
# @click.option('-B', '--buffer', default=16, help='# batches per buffer')
# @click.option('-d', '--dataset', help='relative path of zipped dataset (e.g. Dataset.zip)', required=True)
# def load(batch, buffer, dataset):
#     dataset_path = os.path.abspath(dataset)
#     load_pytorch(batch, buffer, dataset_path)


def main():
    ffysh()



from torch.utils import data
from ..stream.stream import Stream
from ..dataset.dataset import Dataset

class FfyshDataset(data.Dataset):
    def __init__(self, dataset=None):
        if dataset is None:
            raise Exception("No dataset ID was provided.")
        if not isinstance(dataset, Dataset):
            raise Exception("Dataset must be of type Dataset.")
        self.stream = dataset.create_stream()
    
    def __len__(self):
        return self.stream.dataset.length
    
    def __getitem__(self, index):
        if index >= len(self.stream._item_ids):
            next(self.stream.create_iterator(start_index=self.stream._iterator_start_index, batch_size=(index - self.stream._iterator_start_index + 1), pre_download=True))
        return self.stream._item_ids[index]


from ..stream.stream import Stream
from .ffysh_dataset import FfyshDataset
from ..asset.asset import Asset

from torchvision import transforms
from PIL import Image


class ImageDataset(FfyshDataset):
    def __init__(self, dataset=None):
        super().__init__(dataset=dataset)        

    def __len__(self):
        return super().__len__()
    
    def __getitem__(self, index):
        asset_string = super().__getitem__(index)
        asset = self.stream.getitem(item_id=asset_string)
        image = Image.open(asset.path)
        resize = transforms.Resize((640, 640))
        image = resize(image)
        convert_tensor = transforms.ToTensor()
        image_tensor = convert_tensor(image)
        return image_tensor


import json
import sys
import requests
from . import project
from urllib.parse import urljoin

_API_URL = "https://api.flockfysh.ai/"

class BaseUrlSession(requests.Session):
    def __init__(self, base_url=None):
        super().__init__()
        self._authorization = None
        try:
            with open(project.flockfysh_path("credentials.json"), "r") as file:
                self._authorization = json.load(file)["access_token"]
        except (IOError, TypeError, KeyError):
            pass
        self._base_url = base_url

    def request(self, method, url, *args, **kwargs):
        joined_url = urljoin(self._base_url, url)
        if self._authorization:
            headers = {
                **kwargs.get("headers", {}),
                "Authorization": f"Bearer {self._authorization}",
            }
        else:
            headers = kwargs.get("headers", {})
        return super().request(method, joined_url, *args, **kwargs, headers=headers)

    def _authorization_check(self, func):
        def checker(*args, **kwargs):
            if self._authorization is None:
                raise Exception("You need to log in to use Flockfysh CLI.")

            result = self.get("/api/users/auth/").json()
            
            
            if not result["success"] and result["error"]["code"] == "ERROR_UNAUTHORIZED":
                raise Exception("Your Flockfysh CLI credentials are not valid. Please try logging in again.")
            
            elif 'success' not in result:
                raise Exception("Authorization failed")
            
            return func(*args, **kwargs)

        return checker


_api_session = BaseUrlSession(_API_URL)



from .asset import Asset


class Image(Asset):
    def __init__(self, annotation_data: dict, **super_kwargs):
        super().__init__(**super_kwargs)
        self.__annotation_data = annotation_data

    @property
    def annotation_data(self):
        return self.__annotation_data

    def __repr__(self):
        return f"Image {self.id}"

    def _to_dict(self):
        return {
            "_id": self.id,
            "annotation_data": self.annotation_data,
            "path": self._asset_workspace_path,
            "stage": self.stage,
            "mimetype": self.mimetype,
            "type": "image",
        }

    # @property
    # def pytorch_annotation_string(self):
    #     output = []
    #     for annotation_object in self.annotation_data:
    #         output.append(
    #             f'{annotation_object["class"]} {" ".join(str(item) for item in annotation_object["boundingBox"])}')
    #     return "\n".join(output)



import mimetypes
import os
import typing
from ..internals.project import flockfysh_path
import aiofiles
import asyncio
import aiohttp
from ..internals.api import _api_session

#IOModeParameters = typing.Union[str, bytes, os.PathLike[str], os.PathLike[bytes], int, typing.Literal[
 #   "rb+", "r+b", "+rb", "br+", "b+r", "+br", "wb+", "w+b", "+wb", "bw+", "b+w", "+bw", "ab+", "a+b", "+ab", "ba+", "b+a", "+ba", "xb+", "x+b", "+xb", "bx+", "b+x", "+bx", "rb", "br", "rbU", "rUb", "Urb", "brU", "bUr", "Ubr", "wb", "bw", "ab", "ba", "xb", "bx"]]


class Asset:
    def __init__(self, *, asset_id: str, asset_url: str = None, asset_workspace_path: str = None, asset_stage: str,
                 asset_mimetype: str,
                 _internal=False):
        if not _internal:
            raise RuntimeError("You cannot call this class.")

        self.__asset_id = asset_id
        self._asset_workspace_path = asset_workspace_path

        if self._asset_workspace_path is None:
            extension = mimetypes.guess_extension(asset_mimetype)
            if extension:
                abs_path = os.path.join(flockfysh_path("assets"), f"{self.__asset_id}{extension}")
            else:
                abs_path = os.path.join(flockfysh_path("assets"), f"{self.__asset_id}")
            self._asset_workspace_path = os.path.relpath(abs_path, flockfysh_path("."))

        self.__asset_mimetype = asset_mimetype
        self.__asset_url = asset_url
        self.__asset_stage = asset_stage

    async def _download(self):
        if not self.__asset_url:
            try:
                data = _api_session.get(f"/api/datasets/getByAssetId/{self.id}").json()
                self.__asset_url = data["data"]["url"]
            except Exception as e:
                raise RuntimeError("Asset has been deleted. You'll need to create a new stream.")
        os.makedirs(flockfysh_path("assets"), exist_ok=True)
        temp_path = os.path.join(flockfysh_path("assets"), f'{self.__asset_id}.tmp')
        async with aiohttp.ClientSession() as client:
            async with client.get(self.__asset_url) as response:
                if os.path.exists(self.path):
                    return
                async with aiofiles.open(temp_path, "wb") as file:
                    async for chunk, _ in response.content.iter_chunks():
                        await file.write(chunk)
        os.rename(temp_path, self.path)

    @property
    def id(self):
        return self.__asset_id

    @property
    def path(self):
        if not self._asset_workspace_path:
            return None
        else:
            return os.path.join(flockfysh_path("."), self._asset_workspace_path)

    @property
    def is_downloaded(self):
        return os.path.exists(self.path)

    @property
    def mimetype(self):
        return self.__asset_mimetype

    @property
    def stage(self):
        return self.__asset_stage

    @property
    def filename(self):
        return os.path.basename(self.path)

    def free(self, ignore_errors=False):
        try:
            os.remove(self.path)
        except FileNotFoundError:
            return
        except Exception as e:
            if ignore_errors:
                return
            else:
                raise e

    def open(self, mode = "rb", **kwargs):
        if not self.path or not os.path.exists(self.path):
            asyncio.run(self._download())
        return open(self.path, mode=mode, **kwargs)

    async def async_open(self, mode = "rb", **kwargs):
        if not self.path or not os.path.exists(self.path):
            await self._download()
        return aiofiles.open(self.path, mode=mode, **kwargs)

    def _to_dict(self):
        return {
            "_id": self.id,
            "path": self._asset_workspace_path,
            "stage": self.stage,
            "type": "generic",
            "mimetype": self.mimetype,
        }

    def __repr__(self):
        return f"Generic asset {self.__asset_id}"
